@@ -1,14 +1,18 @@
-{-# LANGUAGE DeriveDataTypeable, RecordWildCards, NamedFieldPuns #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE NamedFieldPuns #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE ViewPatterns #-}
 
 ------------------------------------------------------------------------------
 -- |
 -- Module:      Database.PostgreSQL.Simple
 -- Copyright:   (c) 2011 MailRank, Inc.
---              (c) 2011 Leon P Smith
+--              (c) 2011-2012 Leon P Smith
 -- License:     BSD3
 -- Maintainer:  Leon P Smith <leon@melding-monads.com>
 -- Stability:   experimental
--- Portability: portable
 --
 -- A mid-level client library for the PostgreSQL database, aimed at ease of
 -- use and high performance.
@@ -38,6 +42,9 @@ module Database.PostgreSQL.Simple
     -- ** Modifying multiple rows at once
     -- $many
 
+    -- ** @RETURNING@: modifications that returns results
+    -- $returning
+
     -- * Extracting results
     -- $result
 
@@ -51,11 +58,15 @@ module Database.PostgreSQL.Simple
       Base.ConnectInfo(..)
     , Connection
     , Query
+    , ToRow
+    , FromRow
     , In(..)
     , Binary(..)
     , Only(..)
+    , (:.)(..)
     -- ** Exceptions
     , SqlError(..)
+    , PQ.ExecStatus(..)
     , FormatError(fmtMessage, fmtQuery, fmtParams)
     , QueryError(qeMessage, qeQuery)
     , ResultError(errSQLType, errHaskellType, errMessage)
@@ -68,6 +79,9 @@ module Database.PostgreSQL.Simple
     -- * Queries that return results
     , query
     , query_
+    -- ** Queries taking parser as argument
+    , queryWith
+    , queryWith_
     -- * Queries that stream results
     , FoldOptions(..)
     , FetchQuantity(..)
@@ -78,6 +92,7 @@ module Database.PostgreSQL.Simple
     , foldWithOptions_
     , forEach
     , forEach_
+    , returning
     -- * Statements that do not return results
     , execute
     , execute_
@@ -85,18 +100,9 @@ module Database.PostgreSQL.Simple
 --    , Base.insertID
     -- * Transaction handling
     , withTransaction
-    , TransactionMode(..)
-    , IsolationLevel(..)
-    , ReadWriteMode(..)
-    , defaultTransactionMode
-    , defaultIsolationLevel
-    , defaultReadWriteMode
-    , withTransactionLevel
-    , withTransactionMode
+    , withSavepoint
 --    , Base.autocommit
     , begin
-    , beginLevel
-    , beginMode
     , commit
     , rollback
     -- * Helper functions
@@ -107,33 +113,30 @@ module Database.PostgreSQL.Simple
 import           Blaze.ByteString.Builder
                    ( Builder, fromByteString, toByteString )
 import           Blaze.ByteString.Builder.Char8 (fromChar)
+import           Blaze.Text ( integral )
 import           Control.Applicative ((<$>), pure)
-import           Control.Concurrent.MVar
-import           Control.Exception
-                   ( Exception, onException, throw, throwIO, finally )
+import           Control.Exception as E
 import           Control.Monad (foldM)
 import           Data.ByteString (ByteString)
-import           Data.Char(ord)
 import           Data.Int (Int64)
-import qualified Data.IntMap as IntMap
 import           Data.List (intersperse)
-import           Data.Monoid (mappend, mconcat)
+import           Data.Monoid (mconcat)
 import           Data.Typeable (Typeable)
-import           Database.PostgreSQL.Simple.BuiltinTypes
-                   ( oid2builtin, builtin2typname )
-import           Database.PostgreSQL.Simple.Compat ( mask )
+import           Database.PostgreSQL.Simple.Compat ( (<>) )
 import           Database.PostgreSQL.Simple.FromField (ResultError(..))
 import           Database.PostgreSQL.Simple.FromRow (FromRow(..))
 import           Database.PostgreSQL.Simple.Ok
 import           Database.PostgreSQL.Simple.ToField (Action(..), inQuotes)
 import           Database.PostgreSQL.Simple.ToRow (ToRow(..))
 import           Database.PostgreSQL.Simple.Types
-                   ( Binary(..), In(..), Only(..), Query(..) )
+                   ( Binary(..), In(..), Only(..), Query(..), (:.)(..) )
 import           Database.PostgreSQL.Simple.Internal as Base
+import           Database.PostgreSQL.Simple.Transaction
+import           Database.PostgreSQL.Simple.TypeInfo
 import qualified Database.PostgreSQL.LibPQ as PQ
-import           Text.Regex.PCRE.Light (compile, caseless, match)
 import qualified Data.ByteString.Char8 as B
-import qualified Data.Vector as V
+import qualified Data.Text          as T
+import qualified Data.Text.Encoding as TE
 import           Control.Monad.Trans.Reader
 import           Control.Monad.Trans.State.Strict
 
@@ -148,15 +151,6 @@ data FormatError = FormatError {
 
 instance Exception FormatError
 
--- | Exception thrown if 'query' is used to perform an @INSERT@-like
--- operation, or 'execute' is used to perform a @SELECT@-like operation.
-data QueryError = QueryError {
-      qeMessage :: String
-    , qeQuery :: Query
-    } deriving (Eq, Show, Typeable)
-
-instance Exception QueryError
-
 -- | Format a query string.
 --
 -- This function is exposed to help with debugging and logging. Do not
@@ -189,32 +183,144 @@ formatQuery conn q@(Query template) qs
 formatMany :: (ToRow q) => Connection -> Query -> [q] -> IO ByteString
 formatMany _ q [] = fmtError "no rows supplied" q []
 formatMany conn q@(Query template) qs = do
-  case match re template [] of
-    Just [_,before,qbits,after] -> do
+  case parseTemplate template of
+    Just (before, qbits, after) -> do
       bs <- mapM (buildQuery conn q qbits . toRow) qs
       return . toByteString . mconcat $ fromByteString before :
                                         intersperse (fromChar ',') bs ++
                                         [fromByteString after]
-    _ -> error "foo"
-  where
-   re = compile "^([^?]+\\bvalues\\s*)\
-                 \(\\(\\s*[?](?:\\s*,\\s*[?])*\\s*\\))\
-                 \([^?]*)$"
-        [caseless]
+    Nothing -> fmtError "syntax error in multi-row template" q []
 
-escapeStringConn :: Connection -> ByteString -> IO (Maybe ByteString)
-escapeStringConn conn s = withConnection conn $ \c -> do
-   PQ.escapeStringConn c s
+-- Split the input string into three pieces, @before@, @qbits@, and @after@,
+-- following this grammar:
+--
+-- start: ^ before qbits after $
+--     before: ([^?]* [^?\w])? 'VALUES' \s*
+--     qbits:  '(' \s* '?' \s* (',' \s* '?' \s*)* ')'
+--     after:  [^?]*
+--
+-- \s: [ \t\n\r\f]
+-- \w: [A-Z] | [a-z] | [\x80-\xFF] | '_' | '$' | [0-9]
+--
+-- This would be much more concise with some sort of regex engine.
+-- 'formatMany' used to use pcre-light instead of this hand-written parser,
+-- but pcre is a hassle to install on Windows.
+parseTemplate :: ByteString -> Maybe (ByteString, ByteString, ByteString)
+parseTemplate template =
+    -- Convert input string to uppercase, to facilitate searching.
+    search $ B.map toUpper_ascii template
+  where
+    -- Search for the next occurrence of "VALUES"
+    search bs =
+        case B.breakSubstring "VALUES" bs of
+            (x, y)
+                -- If "VALUES" is not present in the string, or any '?' characters
+                -- were encountered prior to it, fail.
+                | B.null y || ('?' `B.elem` x)
+               -> Nothing
+
+                -- If "VALUES" is preceded by an identifier character (a.k.a. \w),
+                -- try the next occurrence.
+                | not (B.null x) && isIdent (B.last x)
+               -> search $ B.drop 6 y
+
+                -- Otherwise, we have a legitimate "VALUES" token.
+                | otherwise
+               -> parseQueryBits $ skipSpace $ B.drop 6 y
+
+    -- Parse '(' \s* '?' \s* .  If this doesn't match
+    -- (and we don't consume a '?'), look for another "VALUES".
+    --
+    -- qb points to the open paren (if present), meaning it points to the
+    -- beginning of the "qbits" production described above.  This is why we
+    -- pass it down to finishQueryBits.
+    parseQueryBits qb
+        | Just ('(', skipSpace -> bs1) <- B.uncons qb
+        , Just ('?', skipSpace -> bs2) <- B.uncons bs1
+        = finishQueryBits qb bs2
+        | otherwise
+        = search qb
+
+    -- Parse (',' \s* '?' \s*)* ')' [^?]* .
+    --
+    -- Since we've already consumed at least one '?', there's no turning back.
+    -- The parse has to succeed here, or the whole thing fails
+    -- (because we don't allow '?' to appear outside of the VALUES list).
+    finishQueryBits qb bs0
+        | Just (')', bs1) <- B.uncons bs0
+        = if '?' `B.elem` bs1
+              then Nothing
+              else Just $ slice3 template qb bs1
+        | Just (',', skipSpace -> bs1) <- B.uncons bs0
+        , Just ('?', skipSpace -> bs2) <- B.uncons bs1
+        = finishQueryBits qb bs2
+        | otherwise
+        = Nothing
+
+    -- Slice a string into three pieces, given the start offset of the second
+    -- and third pieces.  Each "offset" is actually a tail of the uppercase
+    -- version of the template string.  Its length is used to infer the offset.
+    --
+    -- It is important to note that we only slice the original template.
+    -- We don't want our all-caps trick messing up the actual query string.
+    slice3 source p1 p2 =
+        (s1, s2, source'')
+      where
+        (s1, source')  = B.splitAt (B.length source - B.length p1) source
+        (s2, source'') = B.splitAt (B.length p1     - B.length p2) source'
+
+    toUpper_ascii c | c >= 'a' && c <= 'z' = toEnum (fromEnum c - 32)
+                    | otherwise            = c
+
+    -- Based on the definition of {ident_cont} in src/backend/parser/scan.l
+    -- in the PostgreSQL source.  No need to check [a-z], since we converted
+    -- the whole string to uppercase.
+    isIdent c = (c >= '0'    && c <= '9')
+             || (c >= 'A'    && c <= 'Z')
+             || (c >= '\x80' && c <= '\xFF')
+             || c == '_'
+             || c == '$'
+
+    -- Based on {space} in scan.l
+    isSpace_ascii c = (c == ' ') || (c >= '\t' && c <= '\r')
+
+    skipSpace = B.dropWhile isSpace_ascii
+
+escapeStringConn :: Connection -> ByteString -> IO (Either ByteString ByteString)
+escapeStringConn = escapeWrap PQ.escapeStringConn
+
+escapeIdentifier :: Connection -> ByteString -> IO (Either ByteString ByteString)
+escapeIdentifier = escapeWrap PQ.escapeIdentifier
+
+escapeByteaConn :: Connection -> ByteString -> IO (Either ByteString ByteString)
+escapeByteaConn = escapeWrap PQ.escapeByteaConn
+
+escapeWrap       :: (PQ.Connection -> ByteString -> IO (Maybe ByteString))
+                 -> Connection
+                 -> ByteString
+                 -> IO (Either ByteString ByteString)
+escapeWrap f conn s =
+    withConnection conn $ \c ->
+    f c s >>= checkError c
+
+checkError :: PQ.Connection -> Maybe a -> IO (Either ByteString a)
+checkError _ (Just x) = return $ Right x
+checkError c Nothing  = Left . maybe "" id <$> PQ.errorMessage c
 
 buildQuery :: Connection -> Query -> ByteString -> [Action] -> IO Builder
 buildQuery conn q template xs = zipParams (split template) <$> mapM sub xs
-  where quote = inQuotes . fromByteString . maybe undefined id
-        sub (Plain  b) = pure b
-        sub (Escape s) = quote <$> escapeStringConn conn s
-        sub (Many  ys) = mconcat <$> mapM sub ys
+  where err' msg = fmtError (utf8ToString msg) q xs
+        quote = either err' (inQuotes . fromByteString)
+        utf8ToString = T.unpack . TE.decodeUtf8
+        sub (Plain  b)           = pure b
+        sub (Escape s)           = quote <$> escapeStringConn conn s
+        sub (EscapeByteA s)      = quote <$> escapeByteaConn conn s
+        sub (EscapeIdentifier s) = either err' fromByteString <$>
+                                       escapeIdentifier conn s
+        sub (Many  ys)           = mconcat <$> mapM sub ys
         split s = fromByteString h : if B.null t then [] else split (B.tail t)
             where (h,t) = B.break (=='?') s
-        zipParams (t:ts) (p:ps) = t `mappend` p `mappend` zipParams ts ps
+        zipParams (t:ts) (p:ps) = t <> p <> zipParams ts ps
         zipParams [t] []        = t
         zipParams _ _ = fmtError (show (B.count '?' template) ++
                                   " '?' characters, but " ++
@@ -225,67 +331,66 @@ buildQuery conn q template xs = zipParams (split template) <$> mapM sub xs
 --
 -- Returns the number of rows affected.
 --
--- Throws 'FormatError' if the query could not be formatted correctly.
+-- Throws 'FormatError' if the query could not be formatted correctly, or
+-- a 'SqlError' exception if the backend returns an error.
 execute :: (ToRow q) => Connection -> Query -> q -> IO Int64
 execute conn template qs = do
   result <- exec conn =<< formatQuery conn template qs
   finishExecute conn template result
 
--- | A version of 'execute' that does not perform query substitution.
-execute_ :: Connection -> Query -> IO Int64
-execute_ conn q@(Query stmt) = do
-  result <- exec conn stmt
-  finishExecute conn q result
-
 -- | Execute a multi-row @INSERT@, @UPDATE@, or other SQL query that is not
 -- expected to return results.
 --
--- Returns the number of rows affected.
+-- Returns the number of rows affected.   If the list of parameters is empty,
+-- this function will simply return 0 without issuing the query to the backend.
+-- If this is not desired, consider using the 'Values' constructor instead.
 --
--- Throws 'FormatError' if the query could not be formatted correctly.
+-- Throws 'FormatError' if the query could not be formatted correctly, or
+-- a 'SqlError' exception if the backend returns an error.
+--
+-- For example,  here's a command that inserts two rows into a table
+-- with two columns:
+--
+-- @
+-- executeMany c [sql|
+--     INSERT INTO sometable VALUES (?,?)
+--  |] [(1, \"hello\"),(2, \"world\")]
+-- @
+--
+-- Here's an canonical example of a multi-row update command:
+--
+-- @
+-- executeMany c [sql|
+--     UPDATE sometable
+--        SET sometable.y = upd.y
+--       FROM (VALUES (?,?)) as upd(x,y)
+--      WHERE sometable.x = upd.x
+--  |] [(1, \"hello\"),(2, \"world\")
+-- @
+
 executeMany :: (ToRow q) => Connection -> Query -> [q] -> IO Int64
 executeMany _ _ [] = return 0
 executeMany conn q qs = do
   result <- exec conn =<< formatMany conn q qs
   finishExecute conn q result
 
-finishExecute :: Connection -> Query -> PQ.Result -> IO Int64
-finishExecute _conn q result = do
-    status <- PQ.resultStatus result
-    case status of
-      PQ.CommandOk -> do
-          ncols <- PQ.nfields result
-          if ncols /= 0
-          then throwIO $ QueryError ("execute resulted in " ++ show ncols ++
-                                     "-column result") q
-          else do
-            nstr <- PQ.cmdTuples result
-            return $ case nstr of
-                       Nothing  -> 0   -- is this appropriate?
-                       Just str -> toInteger str
-      PQ.TuplesOk -> do
-          ncols <- PQ.nfields result
-          throwIO $ QueryError ("execute resulted in " ++ show ncols ++
-                                 "-column result") q
-      PQ.CopyIn  -> fail "FIXME: postgresql-simple does not currently support COPY IN"
-      PQ.CopyOut -> fail "FIXME: postgresql-simple does not currently support COPY OUT"
-      _ -> do
-        errormsg  <- maybe "" id <$> PQ.resultErrorMessage result
-        statusmsg <- PQ.resStatus status
-        state     <- maybe "" id <$> PQ.resultErrorField result PQ.DiagSqlstate
-        throwIO $ SqlError { sqlState = state
-                           , sqlNativeError = fromEnum status
-                           , sqlErrorMsg = B.concat [ "execute: ", statusmsg
-                                                    , ": ", errormsg ]}
-    where
-     toInteger str = B.foldl' delta 0 str
-                where
-                  delta acc c =
-                    if '0' <= c && c <= '9'
-                    then 10 * acc + fromIntegral (ord c - ord '0')
-                    else error ("finishExecute:  not an int: " ++ B.unpack str)
-
-
+-- | Execute @INSERT ... RETURNING@, @UPDATE ... RETURNING@, or other SQL
+-- query that accepts multi-row input and is expected to return results.
+-- Note that it is possible to write
+--    @'query' conn "INSERT ... RETURNING ..." ...@
+-- in cases where you are only inserting a single row,  and do not need
+-- functionality analogous to 'executeMany'.
+--
+-- If the list of parameters is empty,  this function will simply return @[]@
+-- without issuing the query to the backend.   If this is not desired,
+-- consider using the 'Values' constructor instead.
+--
+-- Throws 'FormatError' if the query could not be formatted correctly.
+returning :: (ToRow q, FromRow r) => Connection -> Query -> [q] -> IO [r]
+returning _ _ [] = return []
+returning conn q qs = do
+  result <- exec conn =<< formatMany conn q qs
+  finishQuery conn q result
 
 -- | Perform a @SELECT@ or other SQL query that is expected to return
 -- results. All results are retrieved and converted before this
@@ -302,17 +407,27 @@ finishExecute _conn q result = do
 --   using 'execute' instead of 'query').
 --
 -- * 'ResultError': result conversion failed.
-query :: (ToRow q, FromRow r)
-         => Connection -> Query -> q -> IO [r]
-query conn template qs = do
-  result <- exec conn =<< formatQuery conn template qs
-  finishQuery conn template result
+--
+-- * 'SqlError':  the postgresql backend returned an error,  e.g.
+--   a syntax or type error,  or an incorrect table or column name.
+query :: (ToRow q, FromRow r) => Connection -> Query -> q -> IO [r]
+query = queryWith fromRow
 
 -- | A version of 'query' that does not perform query substitution.
 query_ :: (FromRow r) => Connection -> Query -> IO [r]
-query_ conn q@(Query que) = do
+query_ = queryWith_ fromRow
+
+-- | A version of 'query' taking parser as argument
+queryWith :: ToRow q => RowParser r -> Connection -> Query -> q -> IO [r]
+queryWith parser conn template qs = do
+  result <- exec conn =<< formatQuery conn template qs
+  finishQueryWith parser conn template result
+
+-- | A version of 'query_' taking parser as argument
+queryWith_ :: RowParser r -> Connection -> Query -> IO [r]
+queryWith_ parser conn q@(Query que) = do
   result <- exec conn que
-  finishQuery conn q result
+  finishQueryWith parser conn q result
 
 -- | Perform a @SELECT@ or other SQL query that is expected to return
 -- results. Results are streamed incrementally from the server, and
@@ -324,6 +439,12 @@ query_ conn q@(Query que) = do
 -- This fold is /not/ strict. The stream consumer is responsible for
 -- forcing the evaluation of its result to avoid space leaks.
 --
+-- This is implemented using a database cursor.    As such,  this requires
+-- a transaction.   This function will detect whether or not there is a
+-- transaction in progress,  and will create a 'ReadCommitted' 'ReadOnly'
+-- transaction if needed.   The cursor is given a unique temporary name,
+-- so the consumer may itself call fold.
+--
 -- Exceptions that may be thrown:
 --
 -- * 'FormatError': the query string could not be formatted correctly.
@@ -332,7 +453,9 @@ query_ conn q@(Query que) = do
 --   using 'execute' instead of 'query').
 --
 -- * 'ResultError': result conversion failed.
-
+--
+-- * 'SqlError':  the postgresql backend returned an error,  e.g.
+--   a syntax or type error,  or an incorrect table or column name.
 fold            :: ( FromRow row, ToRow params )
                 => Connection
                 -> Query
@@ -342,6 +465,9 @@ fold            :: ( FromRow row, ToRow params )
                 -> IO a
 fold = foldWithOptions defaultFoldOptions
 
+-- | Number of rows to fetch at a time.   'Automatic' currently defaults
+--   to 256 rows,  although it might be nice to make this more intelligent
+--   based on e.g. the average size of the rows.
 data FetchQuantity
    = Automatic
    | Fixed !Int
@@ -352,12 +478,19 @@ data FoldOptions
        transactionMode :: !TransactionMode
      }
 
+-- | defaults to 'Automatic',  and 'TransactionMode' 'ReadCommitted' 'ReadOnly'
 defaultFoldOptions :: FoldOptions
 defaultFoldOptions = FoldOptions {
       fetchQuantity   = Automatic,
       transactionMode = TransactionMode ReadCommitted ReadOnly
     }
 
+-- | The same as 'fold',  but this provides a bit more control over
+--   lower-level details.  Currently,  the number of rows fetched per
+--   round-trip to the server and the transaction mode may be adjusted
+--   accordingly.    If the connection is already in a transaction,
+--   then the existing transaction is used and thus the 'transactionMode'
+--   option is ignored.
 foldWithOptions :: ( FromRow row, ToRow params )
                 => FoldOptions
                 -> Connection
@@ -398,7 +531,7 @@ doFold :: ( FromRow row )
        -> a
        -> (a -> row -> IO a)
        -> IO a
-doFold FoldOptions{..} conn _template q a f = do
+doFold FoldOptions{..} conn _template q a0 f = do
     stat <- withConnection conn PQ.transactionStatus
     case stat of
       PQ.TransIdle    -> withTransactionMode transactionMode conn go
@@ -416,11 +549,28 @@ doFold FoldOptions{..} conn _template q a f = do
       PQ.TransUnknown -> fail "foldWithOpts FIXME:  PQ.TransUnknown"
          -- Not sure what this means.
   where
-    go = do
-       -- FIXME:  what about name clashes with already-declared cursors?
-       _ <- execute_ conn ("DECLARE fold NO SCROLL CURSOR FOR "
-                           `mappend` q)
-       loop a `finally` execute_ conn "CLOSE fold"
+    declare = do
+        name <- newTempName conn
+        _ <- execute_ conn $ mconcat
+                 [ "DECLARE ", name, " NO SCROLL CURSOR FOR ", q ]
+        return name
+    fetch (Query name) = query_ conn $
+        Query (toByteString (fromByteString "FETCH FORWARD "
+                             <> integral chunkSize
+                             <> fromByteString " FROM "
+                             <> fromByteString name
+                            ))
+    close name =
+        (execute_ conn ("CLOSE " <> name) >> return ()) `E.catch` \ex ->
+            -- Don't throw exception if CLOSE failed because the transaction is
+            -- aborted.  Otherwise, it will throw away the original error.
+            if isFailedTransactionError ex then return () else throwIO ex
+
+    go = bracket declare close $ \name ->
+         let loop a = do
+                 rs <- fetch name
+                 if null rs then return a else foldM f a rs >>= loop
+          in loop a0
 
 -- FIXME: choose the Automatic chunkSize more intelligently
 --   One possibility is to use the type of the results,  although this
@@ -431,9 +581,6 @@ doFold FoldOptions{..} conn _template q a f = do
     chunkSize = case fetchQuantity of
                  Automatic   -> 256
                  Fixed n     -> n
-    loop a = do
-      rs <- query conn "FETCH FORWARD ? FROM fold" (Only chunkSize)
-      if null rs then return a else foldM f a rs >>= loop
 
 -- | A version of 'fold' that does not transform a state value.
 forEach :: (ToRow q, FromRow r) =>
@@ -463,157 +610,66 @@ forM' lo hi m = loop hi []
            a <- m n
            loop (n-1) (a:as)
 
-finishQuery :: (FromRow r) => Connection -> Query -> PQ.Result -> IO [r]
-finishQuery conn q result = do
+finishQuery :: FromRow r => Connection -> Query -> PQ.Result -> IO [r]
+finishQuery = finishQueryWith fromRow
+
+finishQueryWith :: RowParser r -> Connection -> Query -> PQ.Result -> IO [r]
+finishQueryWith parser conn q result = do
   status <- PQ.resultStatus result
   case status of
+    PQ.EmptyQuery ->
+        throwIO $ QueryError "query: Empty query" q
     PQ.CommandOk -> do
         throwIO $ QueryError "query resulted in a command response" q
     PQ.TuplesOk -> do
-        ncols <- PQ.nfields result
         let unCol (PQ.Col x) = fromIntegral x :: Int
-        typenames <- V.generateM (unCol ncols)
-                                 (\(PQ.Col . fromIntegral -> col) -> do
-                                    getTypename conn =<< PQ.ftype result col)
         nrows <- PQ.ntuples result
         ncols <- PQ.nfields result
         forM' 0 (nrows-1) $ \row -> do
-           let rw = Row row typenames result
-           case runStateT (runReaderT (unRP fromRow) rw) 0 of
+           let rw = Row row result
+           okvc <- runConversion (runStateT (runReaderT (unRP parser) rw) 0) conn
+           case okvc of
              Ok (val,col) | col == ncols -> return val
                           | otherwise -> do
                               vals <- forM' 0 (ncols-1) $ \c -> do
+                                  tinfo <- getTypeInfo conn =<< PQ.ftype result c
                                   v <- PQ.getvalue result row c
-                                  return ( typenames V.! unCol c
+                                  return ( tinfo
                                          , fmap ellipsis v       )
                               throw (ConversionFailed
                                (show (unCol ncols) ++ " values: " ++ show vals)
+                               Nothing
+                               ""
                                (show (unCol col) ++ " slots in target type")
                                "mismatch between number of columns to \
                                \convert and number in target type")
-             Errors []  -> throwIO $ ConversionFailed "" "" "unknown error"
+             Errors []  -> throwIO $ ConversionFailed "" Nothing "" "" "unknown error"
              Errors [x] -> throwIO x
              Errors xs  -> throwIO $ ManyErrors xs
+    PQ.CopyOut ->
+        throwIO $ QueryError "query: COPY TO is not supported" q
+    PQ.CopyIn ->
+        throwIO $ QueryError "query: COPY FROM is not supported" q
+    PQ.BadResponse   -> throwResultError "query" result status
+    PQ.NonfatalError -> throwResultError "query" result status
+    PQ.FatalError    -> throwResultError "query" result status
 
 ellipsis :: ByteString -> ByteString
 ellipsis bs
     | B.length bs > 15 = B.take 10 bs `B.append` "[...]"
     | otherwise        = bs
 
-
--- | Of the four isolation levels defined by the SQL standard,
--- these are the three levels distinguished by PostgreSQL as of version 9.0.
--- See <http://www.postgresql.org/docs/9.1/static/transaction-iso.html>
--- for more information.   Note that prior to PostgreSQL 9.0, 'RepeatableRead'
--- was equivalent to 'Serializable'.
-
-data IsolationLevel
-   = DefaultIsolationLevel  -- ^ the isolation level will be taken from
-                            --   PostgreSQL's per-connection
-                            --   @default_transaction_isolation@ variable,
-                            --   which is initialized according to the
-                            --   server's config.  The default configuration
-                            --   is 'ReadCommitted'.
-   | ReadCommitted
-   | RepeatableRead
-   | Serializable
-     deriving (Show, Eq, Ord, Enum, Bounded)
-
-data ReadWriteMode
-   = DefaultReadWriteMode   -- ^ the read-write mode will be taken from
-                            --   PostgreSQL's per-connection
-                            --   @default_transaction_read_only@ variable,
-                            --   which is initialized according to the
-                            --   server's config.  The default configuration
-                            --   is 'ReadWrite'.
-   | ReadWrite
-   | ReadOnly
-     deriving (Show, Eq, Ord, Enum, Bounded)
-
-data TransactionMode = TransactionMode {
-       isolationLevel :: !IsolationLevel,
-       readWriteMode  :: !ReadWriteMode
-     } deriving (Show, Eq)
-
-defaultTransactionMode :: TransactionMode
-defaultTransactionMode =  TransactionMode
-                            defaultIsolationLevel
-                            defaultReadWriteMode
-
-defaultIsolationLevel  :: IsolationLevel
-defaultIsolationLevel  =  DefaultIsolationLevel
-
-defaultReadWriteMode   :: ReadWriteMode
-defaultReadWriteMode   =  DefaultReadWriteMode
-
--- | Execute an action inside a SQL transaction.
---
--- This function initiates a transaction with a \"@begin
--- transaction@\" statement, then executes the supplied action.  If
--- the action succeeds, the transaction will be completed with
--- 'Base.commit' before this function returns.
---
--- If the action throws /any/ kind of exception (not just a
--- PostgreSQL-related exception), the transaction will be rolled back using
--- 'rollback', then the exception will be rethrown.
-withTransaction :: Connection -> IO a -> IO a
-withTransaction = withTransactionMode defaultTransactionMode
-
--- | Execute an action inside a SQL transaction with a given isolation level.
-withTransactionLevel :: IsolationLevel -> Connection -> IO a -> IO a
-withTransactionLevel lvl
-    = withTransactionMode defaultTransactionMode { isolationLevel = lvl }
-
--- | Execute an action inside a SQL transaction with a given transaction mode.
-withTransactionMode :: TransactionMode -> Connection -> IO a -> IO a
-withTransactionMode mode conn act =
-  mask $ \restore -> do
-    beginMode mode conn
-    r <- restore act `onException` rollback conn
-    commit conn
-    return r
-
--- | Rollback a transaction.
-rollback :: Connection -> IO ()
-rollback conn = execute_ conn "ABORT" >> return ()
-
--- | Commit a transaction.
-commit :: Connection -> IO ()
-commit conn = execute_ conn "COMMIT" >> return ()
-
--- | Begin a transaction.
-begin :: Connection -> IO ()
-begin = beginMode defaultTransactionMode
-
--- | Begin a transaction with a given isolation level
-beginLevel :: IsolationLevel -> Connection -> IO ()
-beginLevel lvl = beginMode defaultTransactionMode { isolationLevel = lvl }
-
--- | Begin a transaction with a given transaction mode
-beginMode :: TransactionMode -> Connection -> IO ()
-beginMode mode conn = do
-    _ <- execute_ conn $! Query (B.concat ["BEGIN", isolevel, readmode])
-    return ()
-  where
-    isolevel = case isolationLevel mode of
-                 DefaultIsolationLevel -> ""
-                 ReadCommitted  -> " READ COMMITTED"
-                 RepeatableRead -> " REPEATABLE READ"
-                 Serializable   -> " SERIALIZABLE"
-    readmode = case readWriteMode mode of
-                 DefaultReadWriteMode -> ""
-                 ReadWrite -> " READ WRITE"
-                 ReadOnly  -> " READ ONLY"
-
 fmtError :: String -> Query -> [Action] -> a
 fmtError msg q xs = throw FormatError {
                       fmtMessage = msg
                     , fmtQuery = q
                     , fmtParams = map twiddle xs
                     }
-  where twiddle (Plain b)  = toByteString b
-        twiddle (Escape s) = s
-        twiddle (Many ys)  = B.concat (map twiddle ys)
+  where twiddle (Plain b)            = toByteString b
+        twiddle (Escape s)           = s
+        twiddle (EscapeByteA s)      = s
+        twiddle (EscapeIdentifier s) = s
+        twiddle (Many ys)            = B.concat (map twiddle ys)
 
 -- $use
 --
@@ -643,7 +699,7 @@ fmtError msg q xs = throw FormatError {
 -- >
 -- > hello = do
 -- >   conn <- connect defaultConnectInfo
--- >   query conn "select 2 + 2"
+-- >   query_ conn "select 2 + 2"
 --
 -- A 'Query' value does not represent the actual query that will be
 -- executed, but is a template for constructing the final query.
@@ -690,7 +746,7 @@ fmtError msg q xs = throw FormatError {
 -- Automated type inference means that you will often be able to avoid
 -- supplying explicit type signatures for the elements of a tuple.
 -- However, sometimes the compiler will not be able to infer your
--- types. Consider a care where you write a numeric literal in a
+-- types. Consider a case where you write a numeric literal in a
 -- parameter tuple:
 --
 -- > query conn "select ? + ?" (40,2)
@@ -712,7 +768,7 @@ fmtError msg q xs = throw FormatError {
 --
 -- The obvious approach would appear to be something like this:
 --
--- > instance (Param a) => QueryParam a where
+-- > instance (ToField a) => ToRow a where
 -- >     ...
 --
 -- Unfortunately, this wreaks havoc with type inference, so we take a
@@ -768,7 +824,7 @@ fmtError msg q xs = throw FormatError {
 -- * There must be no other \"@?@\" characters anywhere in your
 --   template.
 --
--- * There must one or more \"@?@\" in the parentheses.
+-- * There must be one or more \"@?@\" in the parentheses.
 --
 -- * Extra white space is fine.
 --
@@ -789,6 +845,22 @@ fmtError msg q xs = throw FormatError {
 -- > insert into users (first_name,last_name) values
 -- >   ('Boris','Karloff'),('Ed','Wood')
 
+-- $returning
+--
+-- PostgreSQL supports returning values from data manipulation statements
+-- such as @INSERT@ and @UPDATE@.   You can use these statements by
+-- using 'query' instead of 'execute'.   For multi-tuple inserts,
+-- use 'returning' instead of 'executeMany'.
+--
+-- For example, were there an auto-incrementing @id@ column and
+-- timestamp column @t@ that defaulted to the present time for the
+-- @sales@ table, then the following query would insert two new
+-- sales records and also return their new @id@s and timestamps.
+--
+-- > let q = "insert into sales (amount, label) values (?,?) returning id, t"
+-- > xs :: [(Int, UTCTime)] <- query conn q (15,"Sawdust")
+-- > ys :: [(Int, UTCTime)] <- returning conn q [(20,"Chips"),(300,"Wood")]
+
 -- $result
 --
 -- The 'query' and 'query_' functions return a list of values in the
@@ -823,7 +895,7 @@ fmtError msg q xs = throw FormatError {
 --
 -- Although SQL can accommodate @NULL@ as a value for any of these
 -- types, Haskell cannot. If your result contains columns that may be
--- @NULL@, be sure that you use 'Maybe' in those positions of of your
+-- @NULL@, be sure that you use 'Maybe' in those positions of your
 -- tuple.
 --
 -- > (Text, Maybe Int, Int)
@@ -863,25 +935,3 @@ fmtError msg q xs = throw FormatError {
 --   UTF-8. If you use some other encoding, decoding may fail or give
 --   wrong results. In such cases, write a @newtype@ wrapper and a
 --   custom 'Result' instance to handle your encoding.
-
-getTypename :: Connection -> PQ.Oid -> IO ByteString
-getTypename conn@Connection{..} oid =
-  case oid2builtin oid of
-    Just builtin -> return $! builtin2typname builtin
-    Nothing -> modifyMVar connectionObjects $ \oidmap -> do
-      case IntMap.lookup (oid2int oid) oidmap of
-        Just name -> return (oidmap, name)
-        Nothing -> do
-            names <- query conn "SELECT typname FROM pg_type WHERE oid=?"
-                            (Only oid)
-            name <- case names of
-                      []  -> return $ throw SqlError {
-                                         sqlNativeError = -1,
-                                         sqlErrorMsg    = "invalid type oid",
-                                         sqlState       = ""
-                                       }
-                      [Only x] -> return x
-                      _   -> fail "typename query returned more than one result"
-                               -- oid is a primary key,  so the query should
-                               -- never return more than one result
-            return (IntMap.insert (oid2int oid) name oidmap, name)
